import argparse
import pytest
import logging
from tests.helpers_mailman import mailman_create_config, mailman_write_config


for lib in ('urllib3',):
    logger = logging.getLogger(lib)
    logger.setLevel(logging.WARNING)
    logger.propagate = False


@pytest.fixture
def test_options(version, tmpdir):
    api_key = open(f'{version}/apikey').read().strip()
    ip = open(f'{version}/ip').read().strip()
    url = f'http://{ip}'
    n = 'listname'
    mailman_write_config(f'{tmpdir}/config.pck', mailman_create_config())
    return argparse.Namespace(debug=True,
                              list=n,
                              domain='example.com',
                              mailman_config=f'{tmpdir}/config.pck',
                              dry_run=False,
                              api_key=api_key,
                              api_user='api',
                              url=url,
                              mailman_encoding='UTF-8')



import copy
import logging
import os
from packaging import version
import requests
from requests.exceptions import HTTPError
from mailman2discourse.retry import retry

logger = logging.getLogger(__name__)


class DiscourseError(HTTPError):
    pass


class DiscourseClientError(DiscourseError):
    pass


class DiscourseClientNotFound(DiscourseError):
    pass


class DiscourseClientForbidden(DiscourseError):
    pass


DELETE = "DELETE"
GET = "GET"
POST = "POST"
PUT = "PUT"


class DiscourseClient(object):

    def __init__(self, host, api_username, api_key):
        self.host = host
        self.api_username = api_username
        self.api_key = api_key

    def _get(self, path, **kwargs):
        return self._request(GET, path, params=kwargs)

    def _put(self, path, **kwargs):
        return self._request(PUT, path, data=kwargs)

    def _post(self, path, files=None, **kwargs):
        return self._request(POST, path, files=files, params=kwargs)

    def _delete(self, path, **kwargs):
        return self._request(DELETE, path, params=kwargs)

    def _request(
        self, verb, path, params=None, files=None, data=None
    ):
        url = self.host + path

        headers = {
            "Accept": "application/json; charset=utf-8",
            "Api-Key": self.api_key,
            "Api-Username": self.api_username,
            "Discourse-Present": "true",
        }

        request_kwargs = dict(
            allow_redirects=False,
            params=params,
            files=files,
            data=data,
            headers=headers,
        )

        request_message = f"request {verb} {url} params={params} data={data}"
        logger.debug(request_message)
        response = requests.request(verb, url, **request_kwargs)
        logger.debug(f"response {response.status_code}: {repr(response.text[:400])}")

        if response.status_code == 404:
            raise DiscourseClientNotFound(request_message)
        elif response.status_code == 403:
            raise DiscourseClientForbidden(request_message)
        response.raise_for_status()

        if not response.content.strip():
            return None

        json_content = "application/json; charset=utf-8"
        content_type = response.headers["content-type"]
        if content_type != json_content:
            raise DiscourseError(
                'Invalid Response, expecting "{0}" got "{1}"'.format(
                    json_content, content_type
                ),
                response=response,
            )

        try:
            decoded = response.json()
        except ValueError:
            raise DiscourseError("failed to decode response", response=response)

        if isinstance(decoded, dict) and "errors" in decoded:
            message = decoded.get("message")
            if not message:
                message = u",".join(decoded["errors"])
            raise DiscourseClientError(message, response=response)

        return decoded


class DiscourseErrorManyMembers(Exception):
    pass


class DiscourseErrorManyGroupMembers(Exception):
    pass


class DiscourseErrorMinVersion(Exception):
    pass


class DiscourseErrorCategoryFieldType(Exception):
    pass


class DiscourseErrorCategoryNotification(Exception):
    pass


class DiscourseErrorUserCreate(Exception):
    pass


class DiscourseErrorUserActive(Exception):
    pass


class DiscourseErrorUserPreferences(Exception):
    pass


class DiscourseErrorGroupCreate(Exception):
    pass


class DiscourseErrorUserNotFound(Exception):
    pass


class Discourse(object):

    USERNAME_MIN_LENGTH = 3
    USERNAME_MAX_LENGTH = 20

    GROUP_PERMISSIONS_CREATE_REPLY_SEE = 1
    GROUP_PERMISSIONS_SEE = 3

    GROUP_MEMBERS_VISIBILITY_LEVEL_EVERYONE = '0'
    GROUP_MEMBERS_VISIBILITY_LEVEL_MEMBERS = '2'
    GROUP_MEMBERS_VISIBILITY_LEVEL_GROUP_OWNERS_AND_STAFF = '3'

    GROUP_VISIBILITY_LEVEL_EVERYONE = '0'
    GROUP_VISIBILITY_LEVEL_OWNERS_AND_MEMBERS = '2'

    CATEGORY_NOTIFICATION_LEVEL_WATCHING = '3'
    CATEGORY_NOTIFICATION_LEVEL_MUTED = '0'

    GROUP_NAME_LENGTH_MAX = 20

    MIN_VERSION = '2.5.5'

    def __init__(self, args):
        self.args = args

    def connect(self):
        self.d = DiscourseClient(
            self.args.url,
            api_username=self.args.api_user,
            api_key=f'{self.args.api_key}')
        self.about = self.d._get('/about.json')['about']
        logger.info(f'{self.args.url} discourse version {self.about["version"]}')
        return self

    def version_check(self):
        if version.parse(self.about["version"]) < version.parse(Discourse.MIN_VERSION):
            raise DiscourseErrorMinVersion(
                f'discourse is running {self.about["version"]} '
                f'but the minimum required version is {Discourse.MIN_VERSION}')
        return True

    def enable_category_group_moderation_setting(self):
        if version.parse(self.about["version"]) >= version.parse("2.5.9"):
            return 'enable_category_group_moderation'
        else:
            return 'enable_category_group_review'

    @staticmethod
    def value2str(v):
        if v is None:
            return ''
        elif v is True:
            return 'true'
        elif v is False:
            return 'false'
        elif isinstance(v, int):
            return str(v)
        else:
            return v

    @staticmethod
    def category_dict2params(d):
        p = {}
        for k, v in d.items():
            if k not in Discourse.category_fields():
                continue
            if isinstance(v, list):
                if k == 'group_permissions':
                    for v1 in v:
                        p[f'permissions[{v1["group_name"]}]'] = Discourse.value2str(
                            v1["permission_type"])
                else:
                    raise DiscourseErrorCategoryFieldType(
                        f'{k} is not expected to be an array in {d}')
            elif isinstance(v, dict):
                if k == 'custom_fields':
                    for k1, v1 in v.items():
                        p[f'{k}[{k1}]'] = Discourse.value2str(v1)
                else:
                    raise DiscourseErrorCategoryFieldType(
                        f'{k} is not expected to be a dict in {d}')
            else:
                p[k] = Discourse.value2str(v)
        return p

    @staticmethod
    def enforce_group_name_length(name):
        if len(name) > Discourse.GROUP_NAME_LENGTH_MAX:
            shorter = 'g' + name[-(Discourse.GROUP_NAME_LENGTH_MAX - 1):]
            logger.error(
                f'group name {name} length > {Discourse.GROUP_NAME_LENGTH_MAX}'
                f' truncated to {shorter}')
            name = shorter
        return name

    @staticmethod
    def category2moderators_group(name):
        return Discourse.enforce_group_name_length(f'{name}-mo')

    @staticmethod
    def category2members_group(name):
        return Discourse.enforce_group_name_length(f'{name}-me')

    @staticmethod
    def category_fields():
        return (
            'name',
            'id',
            'color',
            'text_color',
            'email_in_allow_strangers',
            'email_in',
            'group_permissions',
            'custom_fields',
            'reviewable_by_group_name',
            'notification_level',
            'topic_url',
            'post_count',
            'topic_count',
        )

    def category_slug(self, name):
        for category in self.d._get('/categories.json')['category_list']['categories']:
            if category['name'] == name:
                return category['slug']
        return None

    def category_get(self, name):
        slug = self.category_slug(name)
        if not slug:
            return None
        try:
            category = self.d._get(f'/c/{slug}/show.json')['category']
            return self.category_dict2params(category)
        except DiscourseClientNotFound:
            return None

    def category_create(self, name):
        category = self.category_get(name)
        if category:
            return category, category
        else:
            if self.args.dry_run:
                return None, {'name': name}
            else:
                kwargs = {
                    'name': name,
                    'color': 'BF1E2E',
                    'text_color': 'FFFFFF',
                    'allow_badges': 'false',
                    # see https://meta.discourse.org/t/79773/157
                    'custom_fields[import_id]': name,
                }
                category = self.d._post('/categories.json', **kwargs)['category']
                return None, self.category_dict2params(category)

    def category_delete(self, name):
        category = self.category_get(name)
        if category and not self.args.dry_run:
            self.d._delete(f'/categories/{category["id"]}')
        return category

    def category_info_get(self, name, post_id=None):
        if not post_id:
            category = self.category_get(name)
            topic_id = os.path.basename(category['topic_url'])
            topic = self.d._get(f'/t/{topic_id}.json')
            post_id = topic['post_stream']['posts'][0]['id']
        post = self.d._get(f'/posts/{post_id}.json')
        return post['raw'], post_id

    def category_info_set(self, name, target):
        before, post_id = self.category_info_get(name)
        if self.args.dry_run:
            return before, target
        self.d._put(f'/posts/{post_id}.json', **{'post[raw]': target})
        return before, self.category_info_get(name, post_id)[0]

    def category_set(self, name, **kwargs):
        before = self.category_get(name)
        target = copy.deepcopy(before)
        #
        # permissions[] obey a precedence algorithm that is non trivial.
        # If the category_set caller specifies permissions[], the existing permissions[*]
        # are discarded. Here is an example why:
        #
        # * the category is set with permissions[everyone] = 1
        # * the caller sets permissions[mygroup] = 1
        # * the CORRECT call is to put
        #
        #   permissions[mygroup] = 1
        #
        #   because it will implicitly discard permissions[everyone]
        #
        # * the INCORRECT call is to put
        #
        #   permissions[mygroup] = 1 and
        #   permissions[everyone] = 1
        #
        #   because permissions[everyone] has precedence and permissions[mygroup] will
        #   be discarded.
        #
        set_permissions = False
        for k in kwargs.keys():
            if k.startswith('permissions['):
                set_permissions = True
        if set_permissions:
            for k in [v for v in target.keys()]:
                if k.startswith('permissions['):
                    del target[k]
        target.update(kwargs)
        if self.args.dry_run:
            return before, target
        else:
            self.d._put(f'/categories/{before["id"]}', **target)
            return before, self.category_get(name)

    def category_notifications(self, email, name, level):
        user = self.user_get(email)
        api_username = self.d.api_username
        try:
            self.d.api_username = user['username'].lower()
            category = self.category_get(name)
            if self.args.dry_run:
                return category['notification_level'], level
            r = self.d._post(f'/category/{category["id"]}/notifications', notification_level=level)
            if not r['success']:
                raise DiscourseErrorCategoryNotification(
                    f'setting notification_level={level} '
                    f'for user {user["username"]} failed: {r["message"]}')
            return category['notification_level'], self.category_get(name)['notification_level']
        finally:
            self.d.api_username = api_username

    def settings_dict2params(self, settings):
        fields = self.settings_fields()
        return {
            s['setting']: s['value'] for s in settings
            if s['setting'] in fields
        }

    def settings_fields(self):
        return ('email_in',
                'log_mail_processing_failures',
                'download_remote_images_to_local',
                'min_post_length'
                'min_first_post_length',
                'min_title_similar_length',
                'disable_system_edit_notifications',
                'disable_emails',
                self.enable_category_group_moderation_setting())

    def settings_set(self, **kwargs):
        before = self.settings_get()
        if self.args.dry_run:
            target = copy.deepcopy(before)
            target.update(kwargs)
            return before, target
        for k, v in kwargs.items():
            self.d._put(f'/admin/site_settings/{k}', **{k: v})
        return before, self.settings_get()

    def settings_get(self):
        return self.settings_dict2params(self.d._get('/admin/site_settings')['site_settings'])

    @staticmethod
    def group_dict2params(group):
        params = {}
        fields = Discourse.group_fields()
        for k, v in group['group'].items():
            if k not in fields:
                continue
            params[f'group[{k}]'] = Discourse.value2str(v)
        return params

    @staticmethod
    def group_fields():
        return ('name',
                'id',
                'public_exit',
                'members_visibility_level',
                'visibility_level',
                'allow_membership_requests',
                'public_admission')

    def group_get(self, name):
        try:
            group = self.d._get(f'/groups/{name}.json')
            return self.group_dict2params(group)
        except DiscourseClientNotFound:
            return None

    def group_create(self, name, **kwargs):
        before = self.group_get(name)
        if before:
            target = copy.deepcopy(before)
        else:
            target = {
                'group[name]': name,
            }
        target.update(kwargs)
        if self.args.dry_run or before == target:
            return before, target
        elif before is None:
            after = self.d._post('/admin/groups', **target)
            after['group'] = after.pop('basic_group')
            return before, self.group_dict2params(after)
        else:
            r = self.d._put(f'/groups/{before["group[id]"]}', **target)
            if not r['success']:
                raise DiscourseErrorGroupCreate(f'{target} failed {r["message"]}')
            return before, self.group_get(name)

    def group_delete(self, name):
        group = self.group_get(name)
        if group and not self.args.dry_run:
            self.d._delete(f'/admin/groups/{group["group[id]"]}')
        return group

    def group_member_fields(self):
        return ('id', 'username', 'owner')

    def group_member_get(self, name, email):
        members = self.d._get(f'/groups/{name}/members.json', filter=email)
        if len(members['members']) == 0:
            return None
        if len(members['members']) > 1:
            raise DiscourseErrorManyGroupMembers(
                f'expected only one member with {email} in group {name} and got {members}')
        if len(members['owners']) > 0:
            member = members['owners'][0]
            member['owner'] = True
        else:
            member = members['members'][0]
            member['owner'] = False
        member['email'] = email
        return {k: member[k] for k in self.group_member_fields()}

    def group_member_create(self, name, email, owner):
        before = self.group_member_get(name, email)

        if not before or self.args.dry_run:
            user = self.user_get(email)
            if not user:
                raise DiscourseErrorUserNotFound(f'{email} is not a know user')

        if self.args.dry_run:
            target = {'email': email, 'username': user['username'], 'owner': owner}
            if before:
                return before, target
            else:
                return None, target

        if not before:
            group_id = self.group_get(name)['group[id]']
            if owner:
                endpoint = f'/admin/groups/{group_id}/owners.json'
                kwargs = {'group[usernames]': user['username']}
            else:
                endpoint = f'/groups/{group_id}/members.json'
                kwargs = {'usernames': user['username']}
            self.d._put(endpoint, **kwargs)
            return None, self.group_member_get(name, email)
        if before['owner'] != owner:
            group_id = self.group_get(name)['group[id]']
            endpoint = f'/admin/groups/{group_id}/owners.json'
            if owner:
                kwargs = {
                    'group[usernames]': before['username'],
                }
                self.d._put(endpoint, **kwargs)
            else:
                self.d._delete(endpoint, user_id=before['id'])
            after = copy.deepcopy(before)
            after['owner'] = owner
            return before, after
        return before, before

    def group_member_delete(self, name, email):
        member = self.group_member_get(name, email)
        if member and not self.args.dry_run:
            group = self.group_get(name)
            self.d._delete(f'/groups/{group["group[id]"]}/members.json', user_id=member['id'])
        return member

    @staticmethod
    def user_dict2params(user):
        return {k: Discourse.value2str(v) for k, v in user.items() if k in Discourse.user_fields()}

    @staticmethod
    def user_fields():
        return ('username', 'name', 'id', 'email', 'active')

    def user_username_exists(self, username):
        try:
            user = self.d._get(f'/users/{username}.json')
            return self.user_dict2params(user['user'])
        except DiscourseClientNotFound:
            return None

    def user_get(self, email):
        email = email.lower()
        users = self.d._get('/admin/users/list/active.json', filter=email, show_emails='true')
        for user in users:
            if user['email'] == email:
                return self.user_dict2params(user)
        return None

    def user_create(self, email, **kwargs):
        before = self.user_get(email)
        if before:
            target = copy.deepcopy(before)
        else:
            target = {'email': email}
        target.update(kwargs)
        if self.args.dry_run or before == target:
            return before, target
        elif before is None:
            r = self.d._post('/users', **target)
            if r['success']:
                return before, self.user_get(email)
            else:
                raise DiscourseErrorUserCreate(f'{target} failed {r["message"]}')
        else:
            self.d._put(f'/u/{before["username"]}', **target)
            after = self.user_get(email)
            if after['active'] != target['active']:
                _, after = self.user_active(email, target['active'])
            return before, after

    def user_active(self, email, active):
        before = self.user_get(email)
        if before['active'] == active:
            return before, before
        if self.args.dry_run:
            after = copy.deepcopy(before)
            after['active'] = active
            return before, after
        if active == 'true':
            action = 'activate'
        elif active == 'false':
            action = 'deactivate'
        else:
            raise DiscourseErrorUserActive(
                f'active={active} invalid, it must be "true" or "false"')
        r = self.d._put(f'/admin/users/{before["id"]}/{action}')
        if r['success']:
            return before, self.user_get(email)
        else:
            raise DiscourseErrorUserActive(f'active={active} failed {r["message"]}')

    def user_delete(self, email):
        user = self.user_get(email)
        if user and not self.args.dry_run:
            self.d._delete(f'/admin/users/{user["id"]}', delete_posts=True)
        return user

    @staticmethod
    def user_preferences_dict2params(user):
        params = {}
        for k, v in user.items():
            if k == 'user_option':
                params.update(Discourse.user_preferences_dict2params(v))
            elif k in Discourse.user_preferences_fields():
                params[k] = Discourse.value2str(v)
        return params

    @staticmethod
    def user_preferences_fields():
        return ('username', 'hide_profile_and_presence', 'locale')

    def user_preferences_get(self, email):
        user = self.user_get(email)
        preferences = self.d._get(f'/u/{user["username"]}.json')
        return self.user_preferences_dict2params(preferences['user'])

    def user_preferences_set(self, email, **kwargs):
        before = self.user_preferences_get(email)
        target = copy.deepcopy(before)
        target.update(kwargs)
        if self.args.dry_run:
            return before, target
        else:
            r = self.d._put(f'/u/{before["username"]}.json', **target)
            if r['success']:
                return before, self.user_preferences_get(email)
            else:
                raise DiscourseErrorUserPreferences(f'{target} failed {r["message"]}')

    def topic_post(self, name, title, message):
        category = self.category_get(name)
        kwargs = {
            'category': str(category['id']),
            'title': title,
            'raw': message,
            'unlist_topic': "false",
            'archetype': "regular",
        }
        return self.d._post('/posts', **kwargs)

    def topic_delete(self, id):
        self.d._delete(f'/t/{id}')

    def message_load(self, raw, msg):
        #
        # There is no convenient way to reverse engineer the /admin/email/handle_mail endpoint
        #
        # https://github.com/discourse/discourse/blob/427d54b2b00fa94474c0522eaed750452c4e7f43/app/controllers/admin/email_controller.rb#L145-L159
        # Which is run by app/jobs/regular/process_email.rb calling Email::Processor.process!
        # Which calls lib/email/processor.rb Email::Receiver.new
        # Which calls lib/email/receiver.rb
        #
        # Errors when processing the queue are found in http://forum.example.com/logs/
        #
        data = {
            'email': raw,
            'api_key': self.d.api_key,
            'api_username': self.d.api_username,
        }
        r = requests.post(f'{self.d.host}/admin/email/handle_mail',
                          allow_redirects=False,
                          data=data)
        logger.debug(r.text)
        r.raise_for_status()
        return r

    def category_topics_delete(self, category_name):
        category = self.category_get(category_name)
        topics = self.d._get(f'/c/{category_name}/{category["id"]}.json')['topic_list']['topics']
        about_id = os.path.basename(category['topic_url'])
        for topic in topics:
            if str(topic['id']) == about_id:
                continue
            self.topic_delete(topic['id'])

    @retry(AssertionError, tries=5)
    def topic_wait(self, category_name, subject):
        topic = self.topic_get(category_name, subject)
        assert topic, f'no topics {subject} in category {category_name}'
        return topic

    def topic_get(self, category_name, subject):
        category = self.category_get(category_name)
        if category is None:
            return None
        kwargs = {
            'term': subject,
            'search_context[type]': 'category',
            'search_context[id]': category["id"],
        }
        r = self.d._get('/search/query', **kwargs)
        if 'topics' not in r:
            return None
        for topic in r['topics']:
            if topic['title'].lower() == subject.lower():
                return topic
        return None

    @retry(AssertionError, tries=5)
    def review_wait(self, subject):
        reviewable = self.reviewable_get(subject)
        assert reviewable, f'{subject} not found in /review'

    def reviewable_get(self, subject):
        reviewables = self.d._get('/review.json')['reviewables']
        for reviewable in reviewables:
            if reviewable['payload']['title'] == subject:
                return reviewable
        return None

    def review_approve(self, title):
        reviewable = self.reviewable_get(title)
        r = self.d._put(f'/review/{reviewable["id"]}/perform/approve_post', version=0)
        r = r['reviewable_perform_result']
        assert r['success']
        return r['created_post_topic_id']

    def disconnect(self):
        del self.d



import argparse
import html
import pytest
from mailman2discourse import mailman
from tests.helpers_mailman import mailman_write_config, mailman_create_config


def test_load_synthetic(tmpdir):
    n = 'listname'
    mailman_write_config(f'{tmpdir}/config.pck', mailman_create_config())
    m = mailman.Mailman(argparse.Namespace(mailman_config=f'{tmpdir}/config.pck',
                                           mailman_encoding='UTF-8',
                                           list=n))
    assert m.load() is True
    assert 'subscribe_policy' in m.info
    assert 'private_roster' in m.info


def test_load_skip(tmpdir, caplog):
    n = 'listname'
    c = mailman_create_config()
    email = 'some@example.com'
    c['digest_members'][email] = 0
    c['language'][email] = 'fr'
    mailman_write_config(f'{tmpdir}/config.pck', c)
    m = mailman.Mailman(argparse.Namespace(mailman_config=f'{tmpdir}/config.pck',
                                           mailman_encoding='UTF-8',
                                           list=n))
    assert m.load() is True
    assert f'SKIP digest_members for unknown user {email}' in caplog.text
    assert f'SKIP language for unknown user {email}' in caplog.text


def test_load_username_none(tmpdir):
    n = 'listname'
    c = mailman_create_config()
    email = 'some@example.com'
    c['members'][email] = 0
    c['usernames'][email] = None
    mailman_write_config(f'{tmpdir}/config.pck', c)
    m = mailman.Mailman(argparse.Namespace(mailman_config=f'{tmpdir}/config.pck',
                                           mailman_encoding='UTF-8',
                                           list=n))
    assert m.load() is True
    assert m.user[email]['usernames'] == ''


@pytest.mark.parametrize("workload,encoding", [('greek', 'iso-8859-1'),
                                               ('regular', 'UTF-8'),
                                               ('instances', 'UTF-8')])
def test_load_realistic(workload, encoding):
    m = mailman.Mailman(argparse.Namespace(mailman_config=f'tests/data/{workload}/config.pck',
                                           mailman_encoding=encoding,
                                           list=workload))
    assert m.load() is True


@pytest.mark.parametrize("raw,cooked", [
    ('my name', 'my name'),
    ('ampersand & not interpreted', 'ampersand & not interpreted'),
    ('&#201; Capital E, acute accent', 'É Capital E, acute accent'),
])
def test_html_unescape(raw, cooked):
    assert html.unescape(raw) == cooked



import pickle


def mailman_create_config():
    c = {
        'description': 'this is a clear description of the category',
        'subscribe_policy': 0,
        'private_roster': 0,
        'archive': 1,
        'archive_private': False,
        # 'default_member_moderation':
        'members': {
        },
        'moderator': [],
        'owner': [],
        'digest_members': {
        },
        'language': {
        },
        'delivery_status': {
        },
        'user_options': {
        },
        'usernames': {
        },
    }
    return c


def mailman_add(c, email, **kwargs):
    for k, v in (('members', True),
                 ('language', 'en'),
                 ('delivery_status', 0),
                 ('user_options', 0),
                 ('usernames', 'My Name')):
        c[k][email] = kwargs.get(k, v)

    for k, v in (('moderator', False),
                 ('owner', False)):
        if kwargs.get(k, v):
            c[k].append(email)


def mailman_write_config(p, c):
    pickle.dump(c, open(p, 'wb'))



import logging
import os
import pickle
import html

logger = logging.getLogger(__name__)


class MailmanUnpickler(pickle.Unpickler):

    class Noop:
        class _BounceInfo:
            pass

    def find_class(self, module, name):
        logger.debug(f'{module} {name}')
        if name == '_Bouncer._BounceInfo':
            return MailmanUnpickler.Noop
        if name == '_BounceInfo':
            return MailmanUnpickler.Noop
        if name == '_Bouncer':
            return MailmanUnpickler.Noop
        return super().find_class(module, name)


class ErrorMailmanNotFound(Exception):
    pass


class Mailman(object):

    MODERATION_HOLD = 0
    MODERATION_REJECT = 1
    MODERATION_DISCARD = 2

    ROSTER_PUBLIC = 0
    ROSTER_MEMBERS = 1
    ROSTER_ADMINS = 2

    SUBSCRIBE_POLICY_OPEN = 0
    SUBSCRIBE_POLICY_CONFIRM = 1
    SUBSCRIBE_POLICY_MODERATE = 2
    SUBSCRIBE_POLICY_CONFIRM_THEN_MODERATE = 3

    def __init__(self, args):
        self.args = args

    @staticmethod
    def moderation_action_mapping(value):
        # Convert the member_moderation_action option to an Action enum.
        # The values were: 0==Hold, 1==Reject, 2==Discard
        return {
            Mailman.MODERATION_HOLD: 'approval',
            Mailman.MODERATION_REJECT: 'ignore',
            Mailman.MODERATION_DISCARD: 'ignore',
            }[value]

    def load_pickle(self):
        p = self.args.mailman_config
        if not os.path.exists(p):
            raise ErrorMailmanNotFound(f'{p} not found, skipping')
        return MailmanUnpickler(open(p, 'rb'), encoding=self.args.mailman_encoding).load()

    def load(self):
        self.name = self.args.list
        c = self.load_pickle()

        logger.debug(f'config.pck is {c}')

        self.info = {
            'private_roster': c['private_roster'],
            'description': c['description'],
            'subscribe_policy': c['subscribe_policy'],
        }
        if c.get('archive'):
            if c.get('archive_private'):
                self.info['archive'] = 'private'
            else:
                self.info['archive'] = 'public'
        else:
            self.info['archive'] = 'private'
        if bool(c.get('default_member_moderation', 0)):
            self.info['moderation'] = self.moderation_action_mapping(
                c['member_moderation_action'])
        else:
            self.info['moderation'] = 'no'
        self.user = {}
        for email in c['members']:
            self.user[email] = {'email': email}
        for k in ('moderator', 'owner'):
            for email in c[k]:
                self.user.setdefault(email, {'email': email})[k] = True
        for email in c['digest_members']:
            if email not in self.user:
                logger.error(f'SKIP digest_members for unknown user {email}')
                continue
            self.user[email]['digest'] = True
        for k in ('language', 'delivery_status', 'user_options'):
            for email, value in c[k].items():
                if email not in self.user:
                    logger.error(f'SKIP {k} for unknown user {email}')
                    continue
                self.user[email][k] = value
        for email, usernames in c['usernames'].items():
            if email not in self.user:
                logger.error(f'SKIP usernames for unknown user {email}')
                continue
            if not isinstance(usernames, str):
                logger.error(f'CONVERT usernames {usernames} for user {email} to an empty string')
                usernames = ''
            self.user[email]['usernames'] = html.unescape(usernames)
        return True



def pytest_addoption(parser):
    parser.addoption(
        "--discourse",
        required=True,
        help="coma separated list of discourse versions v2.5.5,v2.6.0"
    )


#
# https://docs.pytest.org/en/stable/example/parametrize.html
#
def pytest_generate_tests(metafunc):
    if "version" in metafunc.fixturenames:
        versions = metafunc.config.getoption("discourse").split(',')
        metafunc.parametrize("version", versions)



import pytest
import time
from mailman2discourse import discourse


class DiscourseTest(discourse.Discourse):

    def connect(self):
        r = super().connect()
        kwargs = {
            'max_topics_in_first_day': '1000000',
            'max_replies_in_first_day': '1000000',
        }
        self.settings_set(**kwargs)
        return r


def test_connect(test_options):
    d = DiscourseTest(test_options).connect()
    assert 'version' in d.about


def test_enforce_group_name_length():
    good = 'goodname'
    assert good == discourse.Discourse.enforce_group_name_length(good)
    too_long = 'a' * (discourse.Discourse.GROUP_NAME_LENGTH_MAX * 2)
    shorter = discourse.Discourse.enforce_group_name_length(too_long)
    assert len(shorter) == discourse.Discourse.GROUP_NAME_LENGTH_MAX
    assert shorter[0] == 'g'


def test_enable_category_group_moderation_setting(test_options):
    d = DiscourseTest(test_options)
    d.about = {'version': '2.5.5'}
    assert 'group_review' in d.enable_category_group_moderation_setting()
    d.about = {'version': '2.6.0'}
    assert 'group_moderation' in d.enable_category_group_moderation_setting()


def test_version_check(test_options):
    d = DiscourseTest(test_options)
    d.about = {'version': '2.5.5'}
    assert d.version_check() is True
    d.about = {'version': '1.0.0'}
    with pytest.raises(discourse.DiscourseErrorMinVersion):
        d.version_check()


@pytest.mark.parametrize("name", [
    'simpleone',
    'with_underscore',  # _ is converted into -
    'with.dot',  # . is converted into -
])
def test_category_create(test_options, name):
    d = DiscourseTest(test_options).connect()
    try:
        d.args.dry_run = True
        before0, after0 = d.category_create(name)
        assert before0 is None
        assert after0['name'] == name

        d.args.dry_run = False
        before0, after0 = d.category_create(name)
        assert before0 is None
        assert after0['name'] == name
        assert after0['custom_fields[import_id]'] == name

        before1, after1 = d.category_create(name)
        assert after0 == before1
        assert before1 == after1

        d.args.dry_run = True
        assert d.category_delete(name)['id'] == after1['id']
        d.args.dry_run = False
        assert d.category_delete(name)['id'] == after1['id']
        assert d.category_delete(name) is None
    finally:
        d.category_delete(name)


def test_category_set_action(test_options):
    d = DiscourseTest(test_options).connect()
    name = 'CATEGORYNAME'
    group = d.category2members_group(name)
    try:
        _, category = d.category_create(name)
        assert category['name'] == name
        color = '112233'
        before, after = d.category_set(name, color=color)
        assert after['color'] == color

        d.group_create(group)
        permission_group = f'permissions[{group}]'
        permission_one = '1'
        approval = 'custom_fields[require_topic_approval]'
        approval_value = 'true'
        before, after = d.category_set(name, **{
            permission_group: permission_one,
            approval: approval_value,

        })
        assert after[permission_group] == permission_one
        assert after[approval] == approval_value
        assert 'permissions[everyone]' in before
        assert 'permissions[everyone]' not in after

        #
        # permissions[everyone] superseeds permissions[group_name]
        #
        permission_everyone = 'permissions[everyone]'
        before, after = d.category_set(name, **{permission_everyone: permission_one})
        assert after[permission_everyone] == permission_one
        assert permission_group in before
        assert permission_group not in after

    finally:
        d.category_delete(name)
        d.group_delete(group)


def test_category_set_dry_run(test_options):
    d = DiscourseTest(test_options).connect()
    name = 'CATEGORYNAME'
    try:
        _, category = d.category_create(name)
        assert category['name'] == name
        color = '112233'

        d.args.dry_run = True
        before, after = d.category_set(name, color=color)
        assert before['color'] != color
        assert after['color'] == color

        d.args.dry_run = False
        before, after = d.category_set(name, color=color)
        assert before['color'] != color
        assert after['color'] == color

        before, after = d.category_set(name, color=color)
        assert before == after

    finally:
        d.category_delete(name)


def test_group_create(test_options):
    name = 'GROUPNAME'
    try:
        d = DiscourseTest(test_options).connect()

        d.args.dry_run = True
        before0, after0 = d.group_create(name)
        assert before0 is None
        assert after0['group[name]'] == name

        d.args.dry_run = False
        before0, after0 = d.group_create(name, **{'group[public_exit]': 'true'})
        assert before0 is None
        assert after0['group[name]'] == name
        assert after0['group[public_exit]'] == 'true'

        before1, after1 = d.group_create(name, **{'group[public_exit]': 'false'})
        assert after1['group[public_exit]'] == 'false'

        before2, after2 = d.group_create(name)
        assert after1 == before2
        assert before2 == after2

        assert d.group_delete(name)['group[id]'] == after1['group[id]']
        assert d.group_delete(name) is None
    finally:
        d.group_delete(name)


@pytest.mark.parametrize("email", [
    'user@example.com',
    'UPPERCASE@example.com',  # mails are converted to all lowercase
    'some+else@test.com',  # + must be quoted otherwise it fails
])
def test_user_create(test_options, email):
    d = DiscourseTest(test_options).connect()
    try:
        password = 'dashkeedfakojfiesMob'
        username = 'MYUSERNAME'
        kwargs = {
            'username': username,
            'password': password,
        }
        d.args.dry_run = True
        before0, after0 = d.user_create(email, **kwargs)
        assert before0 is None
        assert after0['email'] == email
        assert after0['username'] == username

        d.args.dry_run = False
        before0, after0 = d.user_create(email, **kwargs)
        assert before0 is None
        assert after0['email'] == email.lower()
        assert after0['username'] == username

        before1, after1 = d.user_create(email, **kwargs)
        assert after0 == before1
        assert before1 == after1

        d.args.dry_run = True
        assert d.user_delete(email)['id'] == after1['id']
        d.args.dry_run = False
        assert d.user_delete(email)['id'] == after1['id']
        assert d.user_delete(email) is None
    finally:
        d.user_delete(email)


def test_user_create_active(test_options):
    email = 'user@example.com'
    d = DiscourseTest(test_options).connect()
    try:
        password = 'dashkeedfakojfiesMob'
        username = 'MYUSERNAME'
        kwargs = {
            'username': username,
            'password': password,
            'active': 'false',
        }

        before0, after0 = d.user_create(email, **kwargs)
        assert before0 is None
        assert after0['active'] == 'false'

        kwargs['active'] = 'true'
        before1, after1 = d.user_create(email, **kwargs)
        assert before1['active'] == 'false'
        assert after1['active'] == 'true'

        kwargs['active'] = 'false'
        before2, after2 = d.user_create(email, **kwargs)
        assert before2['active'] == 'true'
        assert after2['active'] == 'false'

    finally:
        d.user_delete(email)


def test_user_active(test_options):
    email = 'user@example.com'
    d = DiscourseTest(test_options).connect()
    try:
        password = 'dashkeedfakojfiesMob'
        username = 'MYUSERNAME'
        kwargs = {
            'username': username,
            'password': password,
            'active': 'true',
        }
        _, _ = d.user_create(email, **kwargs)

        d.args.dry_run = True
        before0, after0 = d.user_active(email, 'false')
        assert before0['active'] == 'true'
        assert after0['active'] == 'false'

        d.args.dry_run = False
        before0, after0 = d.user_active(email, 'false')
        assert before0['active'] == 'true'
        assert after0['active'] == 'false'

        before1, after1 = d.user_active(email, 'true')
        assert before1['active'] == 'false'
        assert after1['active'] == 'true'

        before2, after2 = d.user_active(email, 'true')
        assert after1 == before2
        assert before2 == after2
    finally:
        d.user_delete(email)


def test_user_duplicate(test_options):
    d = DiscourseTest(test_options).connect()
    first_email = 'someuser@some.com'
    second_email = 'user@some.com'
    try:
        password = 'dashkeedfakojfiesMob'
        _, first = d.user_create(first_email, **{
            'active': 'true',
            'username': 'FIRST',
            'password': password,
        })
        _, second = d.user_create(second_email, **{
            'active': 'true',
            'username': 'SECOND',
            'password': password,
        })
        assert first['id'] != second['id']
    finally:
        assert d.user_delete(first_email) is not None
        assert d.user_delete(second_email) is not None


def test_user_name_invalid(test_options):
    d = DiscourseTest(test_options).connect()
    email = 'someuser@some.com'
    with pytest.raises(discourse.DiscourseErrorUserCreate) as e:
        password = 'dashkeedfakojfiesMob'
        invalid_username = 'USERNAME'
        kwargs = {
            'username': invalid_username,
            'password': password,
        }
        _, _ = d.user_create(email, **kwargs)
    assert 'That username is not allowed' in str(e)


def test_user_preferences_set(test_options):
    d = DiscourseTest(test_options).connect()
    email = 'me@example.com'
    try:
        kwargs = {
            'active': 'true',
            'username': 'myusername',
            'password': 'dashkeedfakojfiesMob',
        }
        _, _ = d.user_create(email, **kwargs)
        _, _ = d.user_preferences_set(email, locale='fr', hide_profile_and_presence='true')

        #
        # hide_profile_and_presence is in user['user_option']
        # locale is in user
        #
        d.args.dry_run = True
        before0, after0 = d.user_preferences_set(
            email, locale='en', hide_profile_and_presence='false')
        assert before0['locale'] == 'fr'
        assert before0['hide_profile_and_presence'] == 'true'

        d.args.dry_run = False
        before1, after1 = d.user_preferences_set(
            email, locale='en', hide_profile_and_presence='false')
        assert before0['locale'] == 'fr'
        assert before0['hide_profile_and_presence'] == 'true'
        assert after1['locale'] == 'en'
        assert after1['hide_profile_and_presence'] == 'false'

        before2, after2 = d.user_preferences_set(
            email, locale='en', hide_profile_and_presence='false')
        assert before2 == after2

    finally:
        d.user_delete(email)


def test_group_member(test_options):
    d = DiscourseTest(test_options).connect()
    email = 'user@example.com'
    group_name = 'GROUPNAME'
    try:
        password = 'dashkeedfakojfiesMob'
        username = 'MYUSERNAME'
        _, user = d.user_create(email, **{
            'active': 'true', 'username': username, 'password': password})
        _, group = d.group_create(group_name)

        #
        # Verify idempotence with owner=False
        #
        d.args.dry_run = True
        with pytest.raises(discourse.DiscourseErrorUserNotFound):
            d.group_member_create(group_name, 'unknown@example.com', owner=False)
        before0, after0 = d.group_member_create(group_name, email, owner=False)
        assert before0 is None
        assert after0['username'] == username

        d.args.dry_run = False
        with pytest.raises(discourse.DiscourseErrorUserNotFound):
            d.group_member_create(group_name, 'unknown@example.com', owner=False)
        before0, after0 = d.group_member_create(group_name, email, owner=False)
        assert before0 is None
        assert after0['username'] == username

        before1, after1 = d.group_member_create(group_name, email, owner=False)
        assert after0 == before1
        assert before1 == after1

        #
        # Modify owner to True
        #
        before2, after2 = d.group_member_create(group_name, email, owner=True)
        assert before2['id'] == after2['id']
        assert before2['owner'] is False
        assert after2['owner'] is True
        #
        # Modify owner to False
        #
        before3, after3 = d.group_member_create(group_name, email, owner=False)
        assert before3['id'] == after3['id']
        assert before3['owner'] is True
        assert after3['owner'] is False

        #
        # Delete
        #
        d.args.dry_run = True
        assert d.group_member_delete(group_name, email)['id'] == after1['id']
        d.args.dry_run = False
        assert d.group_member_delete(group_name, email)['id'] == after1['id']
        assert d.group_member_delete(group_name, email) is None

        #
        # Added with owner=True
        #
        before0, after0 = d.group_member_create(group_name, email, owner=True)
        assert before0 is None
        assert after0['username'] == username

    finally:
        d.group_delete(group_name)
        d.user_delete(email)


def test_settings(test_options):
    d = DiscourseTest(test_options).connect()
    saved = d.settings_get()
    try:
        initial_values = {
            'email_in': "true",
            'download_remote_images_to_local': "true",
        }
        d.settings_set(**initial_values)

        kwargs = {
            'email_in': "false",
            'download_remote_images_to_local': "false",
        }
        d.args.dry_run = True
        before0, after0 = d.settings_set(**kwargs)
        assert after0['email_in'] == "false"
        assert after0['download_remote_images_to_local'] == "false"

        d.args.dry_run = False
        before0, after0 = d.settings_set(**kwargs)
        assert after0['email_in'] == "false"
        assert after0['download_remote_images_to_local'] == "false"

        before1, after1 = d.settings_set(**kwargs)
        assert before1 == after0
        assert before1 == after1
    finally:
        d.settings_set(**saved)


def test_category_notifications(test_options):
    email = 'user@example.com'
    category_name = 'CATNAME'
    try:
        d = DiscourseTest(test_options).connect()
        password = 'dashkeedfakojfiesMob'
        username = 'MYUSERNAME'
        _, user = d.user_create(email, **{
            'active': 'true', 'username': username, 'password': password})
        _, category = d.category_create(category_name)

        level0 = '2'
        _, after0 = d.category_notifications(email, category_name, level0)
        assert after0 == level0

        level1 = '1'
        d.args.dry_run = True
        before1, after1 = d.category_notifications(email, category_name, level1)
        assert before1 == after0
        assert after1 == level1

        d.args.dry_run = False
        before1, after1 = d.category_notifications(email, category_name, level1)
        assert before1 == after0
        assert after1 == level1

        before2, after2 = d.category_notifications(email, category_name, level1)
        assert before2 == after1
        assert before2 == after2
    finally:
        d.user_delete(email)
        d.category_delete(category_name)


def test_category_info(test_options):
    d = DiscourseTest(test_options).connect()
    name = 'ABCD'
    try:
        _, _ = d.category_create(name)
        original = 'This is the original content to look for'
        _, _ = d.category_info_set(name, original)

        d.args.dry_run = True
        modified = 'Something different is said about the category'
        before0, after0 = d.category_info_set(name, modified)
        assert before0 == original
        assert after0 == modified

        d.args.dry_run = False
        before0, after0 = d.category_info_set(name, modified)
        assert before0 == original
        assert after0 == modified

        before1, after1 = d.category_info_set(name, modified)
        assert after0 == before1
        assert before1 == after1

    finally:
        d.category_delete(name)


def test_message_load(test_options):
    topic = None
    email = 'user@example.com'
    category_name = 'CATNAME'
    try:
        d = DiscourseTest(test_options).connect()
        password = 'dashkeedfakojfiesMob'
        username = 'MYUSERNAME'
        _, user = d.user_create(email, **{
            'active': 'true', 'username': username, 'password': password})
        _, category = d.category_create(category_name)
        d.settings_set(email_in='true',
                       log_mail_processing_failures='true',
                       email_in_min_trust='0',
                       min_post_length='5',
                       min_first_post_length='5',
                       min_title_similar_length='10240')
        email_in = 'list@example.com'
        d.category_set(category_name,
                       email_in=email_in,
                       email_in_allow_strangers='true')
        subject = f'MESSAGE ONE {time.time()}'
        message = f"""From user@example.com  Mon Nov  9 21:54:11 1999
Message-ID: <msg@{time.time()}>
From: "Some One" <user@example.com>
To: <list@example.com>
Date: Fri,  6 Apr 2007 15:43:55 -0700 (PDT)
Subject: {subject}

First content
        """
        assert 'email has been received' in d.message_load(message, None).text
        topic = d.topic_wait(category_name, subject)
    finally:
        if topic:
            d.topic_delete(topic['id'])
        d.user_delete(email)
        d.category_delete(category_name)
        d.settings_set(email_in='false')



import argparse
import logging
import sys
from mailman2discourse.importer import Importer

logging.basicConfig(format='%(funcName)s %(levelname)s %(message)s')


class Cmd(object):

    def __init__(self, *args):
        self.args = self.parser().parse_args(args)
        self.configure_logging()

    def configure_logging(self):
        logger = logging.getLogger('')
        logger.setLevel(logging.WARNING)

        logger = logging.getLogger('mailman2discourse')
        if self.args.debug:
            logger.setLevel(logging.DEBUG)
        elif self.args.verbose:
            logger.setLevel(logging.INFO)
        else:
            logger.setLevel(logging.WARNING)
        return logger.level

    def parser(self):
        parser = argparse.ArgumentParser(
            description="")

        parser.add_argument('--api-key',
                            help='Discourse api key',
                            required=True)
        parser.add_argument('--api-user',
                            help='Discourse api user',
                            default='api')
        parser.add_argument('--url',
                            help='Discourse URL',
                            required=True)
        parser.add_argument('--domain',
                            help='domain name of the mailing list',
                            required=True)
        parser.add_argument('--list',
                            help='name of the mailing list and the discourse category',
                            required=True)
        parser.add_argument('--mailman-config',
                            help='path to the config.pck file',
                            required=True)
        parser.add_argument('--mailman-encoding',
                            help='encoding of the config.pck file',
                            default='UTF-8')
        parser.add_argument('--verbose', action='store_const',
                            const=True,
                            help='enable verbose logging')
        parser.add_argument('--debug', action='store_const',
                            const=True,
                            help='enable debug logging')
        parser.add_argument('--dry-run', action='store_const',
                            const=True,
                            help='display what would be done but do nothing')
        return parser

    def main(self):
        return Importer(self.args).main()


def main(argv=sys.argv[1:]):
    return Cmd(*argv).main()



import copy
from deepdiff import DeepDiff
import hashlib
import logging
import os
from mailman2discourse.mailman import Mailman
from mailman2discourse.discourse import Discourse
import re
import unidecode

logger = logging.getLogger(__name__)


class Importer(object):

    def __init__(self, args):
        self.args = args
        self.discourse = Discourse(self.args)
        self.discourse.connect()
        self.mailman = Mailman(self.args)
        self.mailman.load()

    def name_get(self):
        return self.mailman.name

    def email_get(self):
        return f'{self.name_get()}@{self.args.domain}'

    def members_get(self):
        return self.mailman.user.values()

    def owners_get(self):
        return [u for u in self.mailman.user.values() if u.get('owner')]

    def moderators_get(self):
        return [u for u in self.mailman.user.values() if u.get('moderator')]

    def group_moderators_name(self):
        return Discourse.category2moderators_group(self.name_get())

    def group_members_name(self):
        return Discourse.category2members_group(self.name_get())

    def users_add(self):
        before = {}
        after = {}
        for email in [e for e in self.mailman.user.keys()]:
            user = self.mailman.user[email]
            name = user.get('usernames', '')
            username = None
            if name:
                candidate = re.sub(r'^[\W_]*(.*?)[\W_]*$', r'\1', unidecode.unidecode(name))
                candidate = re.sub(r'[\W_]+', r'_', candidate)
                too_short = len(candidate) < Discourse.USERNAME_MIN_LENGTH
                too_long = len(candidate) > Discourse.USERNAME_MAX_LENGTH
                if (not too_short and not too_long and not self.discourse.user_username_exists(
                        candidate)):
                    username = candidate
            if not username:
                username = hashlib.md5(email.encode('utf-8')).hexdigest()[:20]
            password = hashlib.md5(os.urandom(16)).hexdigest()[:20]
            try:
                before[email], after[email] = self.discourse.user_create(email, **{
                    'username': username,
                    'password': password,
                    'active': 'true',
                    'name': name})
            except Exception as e:
                logger.error(f'SKIP user {user} because {e}')
                del self.mailman.user[email]
        return before, after

    def users_preferences(self):
        before = {}
        after = {}
        default_preferences = {}
        if self.mailman.info['private_roster'] != Mailman.ROSTER_PUBLIC:
            default_preferences['hide_profile_and_presence'] = 'true'
        for email, user in self.mailman.user.items():
            preferences = copy.deepcopy(default_preferences)
            language = user.get('language')
            if language:
                preferences['locale'] = language
            if preferences != {}:
                before[email], after[email] = self.discourse.user_preferences_set(
                    email, **preferences)
        return before, after

    def users_notifications(self):
        before = {}
        after = {}
        for email in [e for e in self.mailman.user.keys()]:
            (before[email], after[email]) = self.discourse.category_notifications(
                email, self.name_get(), self.notification_level_get(email))
        return before, after

    def group_members_add(self, group_name, members, owners):
        before = {}
        after = {}
        for users, owner in ((owners, True), (members, False)):
            for user in users:
                email = user['email']
                if email in after:  # owners take precendence over members
                    continue
                before[email], after[email] = self.discourse.group_member_create(
                    group_name, email, owner=owner)
        return before, after

    def notification_level_get(self, email):
        if self.mailman.user[email].get('delivery_status') in (0, '', None):
            return Discourse.CATEGORY_NOTIFICATION_LEVEL_WATCHING
        else:
            return Discourse.CATEGORY_NOTIFICATION_LEVEL_MUTED

    def moderation_get(self):
        return self.mailman.info['moderation']

    def category_settings(self):
        kwargs = {
            'email_in': self.email_get(),
        }

        create_reply_see = Discourse.GROUP_PERMISSIONS_CREATE_REPLY_SEE
        see = Discourse.GROUP_PERMISSIONS_SEE
        archive = self.mailman.info['archive']

        if self.moderation_get() == 'no' and archive == 'public':
            kwargs.update({
                'email_in_allow_strangers': 'true',
                'permissions[everyone]': create_reply_see,
            })
        elif self.moderation_get() == 'no' and archive == 'private':
            kwargs.update({
                'email_in_allow_strangers': 'true',
                f'permissions[{self.group_members_name()}]': create_reply_see,
            })
        elif self.moderation_get() == 'ignore' and archive == 'public':
            kwargs.update({
                'email_in_allow_strangers': 'false',
                f'permissions[{self.group_members_name()}]': create_reply_see,
                'permissions[everyone]': see,
            })
        elif self.moderation_get() == 'ignore' and archive == 'private':
            kwargs.update({
                'email_in_allow_strangers': 'false',
                f'permissions[{self.group_members_name()}]': create_reply_see,
            })
        elif self.moderation_get() == 'approval' and archive == 'public':
            kwargs.update({
                'email_in_allow_strangers': 'true',
                'custom_fields[require_topic_approval]': 'true',
                'custom_fields[require_reply_approval]': 'true',
                'reviewable_by_group_name': self.group_moderators_name(),
            })
        elif self.moderation_get() == 'approval' and archive == 'private':
            kwargs.update({
                'email_in_allow_strangers': 'true',
                f'permissions[{self.group_members_name()}]': create_reply_see,
                'custom_fields[require_topic_approval]': 'true',
                'custom_fields[require_reply_approval]': 'true',
                'reviewable_by_group_name': self.group_moderators_name(),
            })
        else:
            raise Exception(f'unexpected combination moderation={self.moderation_get()} ',
                            f'and archive=={archive}')
        logger.info(kwargs)
        return kwargs

    def group_moderators_settings(self):
        members_visibility_level = {
            Mailman.ROSTER_PUBLIC: Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_EVERYONE,
            Mailman.ROSTER_MEMBERS: Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_MEMBERS,
            Mailman.ROSTER_ADMINS: Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_GROUP_OWNERS_AND_STAFF,
        }[self.mailman.info['private_roster']]
        return {
            'group[visibility_level]': '0',
            'group[members_visibility_level]': members_visibility_level,
            'group[public_admission]': 'false',
            'group[allow_membership_requests]': 'false',
        }

    def group_members_settings(self):
        private_roster = self.mailman.info['private_roster']
        subscribe_policy = self.mailman.info['subscribe_policy']
        if subscribe_policy in (Mailman.SUBSCRIBE_POLICY_OPEN,
                                Mailman.SUBSCRIBE_POLICY_CONFIRM):
            policy = 'public'
        elif subscribe_policy in (Mailman.SUBSCRIBE_POLICY_MODERATE,
                                  Mailman.SUBSCRIBE_POLICY_CONFIRM_THEN_MODERATE):
            policy = 'moderate'
        else:
            logger.error(f'subscribe_policy is unknown ({subscribe_policy} fallback to 2/moderate')
            policy = 'moderate'

        m_everyone = Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_EVERYONE
        m_members = Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_MEMBERS
        m_group_owners_and_staff = Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_GROUP_OWNERS_AND_STAFF
        g_everyone = Discourse.GROUP_VISIBILITY_LEVEL_EVERYONE
        g_owners_and_members = Discourse.GROUP_VISIBILITY_LEVEL_OWNERS_AND_MEMBERS
        converter = {
            (Mailman.ROSTER_PUBLIC, 'public'): {
                'group[public_admission]': 'true',
                'group[allow_membership_requests]': 'false',
                'group[visibility_level]': g_everyone,
                'group[members_visibility_level]': m_everyone,
            },
            (Mailman.ROSTER_MEMBERS, 'public'): {
                'group[public_admission]': 'true',
                'group[allow_membership_requests]': 'false',
                'group[visibility_level]': g_everyone,
                'group[members_visibility_level]': m_members,
            },
            (Mailman.ROSTER_ADMINS, 'public'): {
                'group[public_admission]': 'true',
                'group[allow_membership_requests]': 'false',
                'group[visibility_level]': g_owners_and_members,
                'group[members_visibility_level]': m_group_owners_and_staff,
            },
            (Mailman.ROSTER_PUBLIC, 'moderate'): {
                'group[public_admission]': 'false',
                'group[allow_membership_requests]': 'true',
                'group[visibility_level]': g_everyone,
                'group[members_visibility_level]': m_everyone,
            },
            (Mailman.ROSTER_MEMBERS, 'moderate'): {
                'group[public_admission]': 'false',
                'group[allow_membership_requests]': 'true',
                'group[visibility_level]': g_everyone,
                'group[members_visibility_level]': m_members,
            },
            (Mailman.ROSTER_ADMINS, 'moderate'): {
                'group[public_admission]': 'false',
                'group[allow_membership_requests]': 'false',
                'group[visibility_level]': g_owners_and_members,
                'group[members_visibility_level]': m_group_owners_and_staff,
            },
        }
        return converter[(private_roster, policy)]

    def settings(self):
        kwargs = {
            'disallow_reply_by_email_after_days': '0',
            'block_auto_generated_emails': 'false',
            'automatically_download_gravatars': 'false',
            'body_min_entropy': '0',
            'title_min_entropy': '0',
            'allow_uppercase_posts': 'true',
            'title_prettify': 'false',
            'allow_user_locale': 'true',
            'email_in': 'true',
            'log_mail_processing_failures': 'true',
            'download_remote_images_to_local': 'true',
            'min_post_length': '1',
            'min_first_post_length': '1',
            'min_title_similar_length': '10240',
            'disable_system_edit_notifications': 'true',
            'hide_user_profiles_from_public': 'true',
            self.discourse.enable_category_group_moderation_setting(): 'true',
        }
        return self.discourse.settings_set(**kwargs)

    def importer(self):
        before = {
        }
        after = {
        }
        (before['settings'], after['settings']) = self.settings()
        (before['category'], _) = self.discourse.category_create(self.name_get())
        if len(self.mailman.info['description']) > 0:
            (before['category-info'], after['category-info']) = self.discourse.category_info_set(
                self.name_get(), self.mailman.info['description'])
        logger.info('Adding users')
        (before['members'], after['members']) = self.users_add()
        logger.info('Populating the moderators group')
        (before['group-moderators'], after['group-moderators']) = (
            self.discourse.group_create(self.group_moderators_name()))
        (before['group-moderators-list'], after['group-moderators-list']) = self.group_members_add(
            self.group_moderators_name(), self.moderators_get(), self.owners_get())
        #
        # The settings must be modified after members are added because some of them
        # can only be set if the group has at least one owner.
        #
        (_, after['group-moderators']) = self.discourse.group_create(
            self.group_moderators_name(), **self.group_moderators_settings())

        if self.moderation_get() != 'no' or self.mailman.info['archive'] == 'private':
            logger.info('Populating the members group')
            (before['group-members'], _) = self.discourse.group_create(self.group_members_name())
            (before['group-members-list'],
             after['group-members-list']) = self.group_members_add(
                 self.group_members_name(), self.members_get(), self.owners_get())
            (_, after['group-members']) = self.discourse.group_create(
                self.group_members_name(), **self.group_members_settings())

        (_, after['category']) = self.discourse.category_set(
            self.name_get(), **self.category_settings())

        (before['preferences'], after['preferences']) = self.users_preferences()
        (before['notifications'], after['notifications']) = self.users_notifications()

        return before, after

    def main(self):
        before, after = self.importer()
        print(DeepDiff(before, after).pretty())
        return 0



import filecmp
import os


def test_requirements(tmpdir):
    requirements = f'{tmpdir}/r.txt'
    requirements_dev = f'{tmpdir}/r-dev.txt'
    os.system('pipenv run pipenv_to_requirements -f '
              f'-o {requirements} -d {requirements_dev}')
    r = filecmp.cmp('requirements.txt', requirements, shallow=False)
    r_dev = filecmp.cmp('requirements-dev.txt', requirements_dev, shallow=False)
    assert r and r_dev, ('requirements.txt needs updating from Pipfile, run'
                         ' pipenv run pipenv_to_requirements -f')



import logging
from mailman2discourse.cmd import Cmd
import pytest


@pytest.mark.parametrize("args,level", [([], logging.WARNING),
                                        (['--verbose'], logging.INFO),
                                        (['--debug'], logging.DEBUG)])
def test_logging(args, level):
    args += [
        '--api-key=KEY', '--url=URL',
        '--list=LIST', '--domain=example.com',
        '--mailman-config=PATH',
    ]
    cmd = Cmd(*args)
    assert cmd.configure_logging() == level



from mailman2discourse import importer
from mailman2discourse import discourse
from mailman2discourse.discourse import Discourse
from mailman2discourse.mailman import Mailman
from mailman2discourse import retry
import pytest
from tests.helpers_mailman import (
    mailman_write_config,
    mailman_add)
import time


def test_main(test_options, mocker):
    m = mocker.patch('mailman2discourse.importer.Importer.importer',
                     return_value=(None, None))
    i = importer.Importer(test_options)
    assert i.main() == 0
    m.assert_called()


def test_group_moderators_settings(test_options):
    i = importer.Importer(test_options)
    i.mailman.info['private_roster'] = 2
    s = i.group_moderators_settings()
    assert s['group[members_visibility_level]'] == '3'


def test_group_members_settings(test_options):
    m_everyone = Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_EVERYONE
    g_everyone = Discourse.GROUP_VISIBILITY_LEVEL_EVERYONE

    incoming = (Mailman.ROSTER_PUBLIC, Mailman.SUBSCRIBE_POLICY_OPEN)
    expected = {
         'group[public_admission]': 'true',
         'group[allow_membership_requests]': 'false',
         'group[visibility_level]': g_everyone,
         'group[members_visibility_level]': m_everyone,
     }

    i = importer.Importer(test_options)
    (i.mailman.info['private_roster'], i.mailman.info['subscribe_policy']) = incoming
    assert i.group_members_settings() == expected


def test_users_add_invalid(test_options):
    i = importer.Importer(test_options)
    invalid_email = 'invalidmail@'
    valid_email = 'valid@example.com'
    try:
        i.mailman.user = {
            invalid_email: {},
            valid_email: {},
        }
        before, after = i.users_add()
        assert before == {valid_email: None}
        assert invalid_email not in after
        assert valid_email in after
        assert invalid_email not in i.mailman.user
    finally:
        i.discourse.user_delete(invalid_email)
        i.discourse.user_delete(valid_email)


@pytest.mark.parametrize("raw,cooked", [
    ('Loïc Dachary', 'Loic_Dachary'),
    ('  - / my name _.', 'my_name'),
    ('__other name__', 'other_name'),
    ('one__underscore', 'one_underscore'),
    ('a hash is used if the name is too long', 'b58996c504c5638798eb'),
])
def test_users_add_normalize(test_options, raw, cooked):
    i = importer.Importer(test_options)
    email = 'user@example.com'
    try:
        i.mailman.user = {
            email: {
                'usernames': raw,
            },
        }
        before, after = i.users_add()
        assert after[email]['username'] == cooked
    finally:
        i.discourse.user_delete(email)


def test_importer_moderation_no(test_options):
    i = importer.Importer(test_options)
    owner = 'owner@example.com'
    moderator = 'moderator@example.com'

    def cleanup():
        i.discourse.group_delete(i.group_moderators_name())
        i.discourse.category_delete(i.name_get())
        i.discourse.user_delete(moderator)
        i.discourse.user_delete(owner)
    cleanup()
    try:
        c = i.mailman.load_pickle()
        mailman_add(c, owner, owner=True)
        mailman_add(c, moderator, moderator=True)
        mailman_write_config(f'{test_options.mailman_config}', c)
        i.mailman.load()

        before0, after0 = i.importer()
        assert before0['category'] is None
        assert list(before0['group-moderators-list'].keys()) == [owner, moderator]
        assert all(v is None for v in before0['group-moderators-list'].values())
        assert 'group-members-list' not in before0
        assert before0['group-moderators'] is None
        assert list(before0['members'].keys()) == [owner, moderator]
        assert 'email_in' in before0['settings']

        assert after0['category']['name'] == test_options.list
        assert after0['settings']['email_in'] == 'true'
        assert after0['group-moderators']['group[name]'].startswith(test_options.list)
        assert owner in after0['group-moderators-list']
        assert moderator in after0['group-moderators-list']
        assert 'group-members' not in after0

        before1, after1 = i.importer()
        assert before1 == after0
        assert before1 == after1
    finally:
        cleanup()


def test_importer_moderation_approval(test_options):
    i = importer.Importer(test_options)
    owner = 'owner@example.com'
    moderator = 'moderator@example.com'
    member = 'member@example.com'
    try:
        c = i.mailman.load_pickle()
        c['default_member_moderation'] = True
        c['member_moderation_action'] = Mailman.MODERATION_HOLD
        mailman_add(c, owner, owner=True)
        mailman_add(c, moderator, moderator=True)
        mailman_add(c, member)
        mailman_write_config(test_options.mailman_config, c)
        i.mailman.load()

        before0, after0 = i.importer()
        assert before0['category'] is None
        assert list(before0['group-moderators-list'].keys()) == [owner, moderator]
        assert all(v is None for v in before0['group-moderators-list'].values())
        assert list(before0['group-members-list'].keys()) == [owner, moderator, member]
        assert all(v is None for v in before0['group-members-list'].values())
        assert before0['group-moderators'] is None
        assert list(before0['members'].keys()) == [owner, moderator, member]
        assert 'email_in' in before0['settings']

        assert after0['category']['name'] == test_options.list
        assert after0['settings']['email_in'] == 'true'
        assert after0['group-moderators']['group[name]'].startswith(test_options.list)
        for who in (owner, member, moderator):
            assert who in after0['group-members-list']
        for who in (owner, moderator):
            assert who in after0['group-moderators-list']
        assert member not in after0['group-moderators-list']

        before1, after1 = i.importer()
        assert before1 == after0
        assert before1 == after1
    finally:
        i.discourse.group_delete(i.group_moderators_name())
        i.discourse.group_delete(i.group_members_name())
        i.discourse.category_delete(i.name_get())
        i.discourse.user_delete(moderator)
        i.discourse.user_delete(owner)
        i.discourse.user_delete(member)


def test_importer_members_settings(test_options):
    i = importer.Importer(test_options)
    owner = 'owner@example.com'
    moderator = 'moderator@example.com'
    member = 'member@example.com'
    try:
        c = i.mailman.load_pickle()
        c['default_member_moderation'] = True
        c['member_moderation_action'] = Mailman.MODERATION_HOLD
        c['private_roster'] = Mailman.ROSTER_ADMINS
        c['subscribe_policy'] = Mailman.SUBSCRIBE_POLICY_MODERATE
        mailman_add(c, owner, owner=True)
        mailman_add(c, moderator, moderator=True)
        mailman_add(c, member)
        mailman_write_config(test_options.mailman_config, c)
        i.mailman.load()

        before0, after0 = i.importer()
        assert before0['group-members'] is None
        assert (after0['group-members']['group[visibility_level]'] ==
                Discourse.GROUP_VISIBILITY_LEVEL_OWNERS_AND_MEMBERS)
        assert (after0['group-members']['group[members_visibility_level]'] ==
                Discourse.GROUP_MEMBERS_VISIBILITY_LEVEL_GROUP_OWNERS_AND_STAFF)

        before1, after1 = i.importer()
        assert before1 == after0
        assert before1 == after1
    finally:
        i.discourse.group_delete(i.group_moderators_name())
        i.discourse.group_delete(i.group_members_name())
        i.discourse.category_delete(i.name_get())
        i.discourse.user_delete(moderator)
        i.discourse.user_delete(owner)
        i.discourse.user_delete(member)


def email_stranger(d, email):
    subject = f'MESSAGE {time.time()}'
    message = f"""From user@example.com  Mon Nov  9 21:54:11 1999
Message-ID: <msg@{time.time()}>
From: "Some One" <stranger@example.com>
To: <{email}>
Date: Fri,  6 Apr 2007 15:43:55 -0700 (PDT)
Subject: {subject}

First content
    """
    assert 'email has been received' in d.message_load(message, None).text
    return subject


def email_stranger_ok(i, email):
    subject = email_stranger(i.discourse, email)
    topic_id = i.discourse.topic_wait(i.name_get(), subject)['id']
    return subject, topic_id


def email_stranger_ignored(i, email):
    subject = email_stranger(i.discourse, email)
    with pytest.raises(retry.RetryException):
        i.discourse.topic_wait(i.name_get(), subject)


def email_stranger_and_approve(i, email):
    subject = email_stranger(i.discourse, email)
    i.discourse.review_wait(subject)
    return subject, i.discourse.review_approve(subject)


def topic_post(i, email):
    title = f'a perfectly fine msg from {email} {time.time()}'
    message = f'thanks {email} for a nice and interesting message {time.time()}'
    user = i.discourse.user_get(email)
    api_username = i.discourse.d.api_username
    try:
        i.discourse.d.api_username = user['username']
        return title, i.discourse.topic_post(i.name_get(), title, message)
    finally:
        i.discourse.d.api_username = api_username


def topic_post_ok(i, email):
    title, r = topic_post(i, email)
    return title, r['topic_id']


def topic_post_forbidden(i, email):
    with pytest.raises(discourse.DiscourseClientForbidden):
        topic_post(i, email)


def topic_post_and_approve(i, email):
    title, r = topic_post(i, email)
    assert r is None
    return title, i.discourse.review_approve(title)


def user_can_see_topic(i, email, subject):
    user = i.discourse.user_get(email)
    api_username = i.discourse.d.api_username
    try:
        i.discourse.d.api_username = user['username']
        return i.discourse.topic_wait(i.name_get(), subject)['id']
    finally:
        i.discourse.d.api_username = api_username


def user_cannot_see_category(i, email):
    user = i.discourse.user_get(email)
    api_username = i.discourse.d.api_username
    try:
        i.discourse.d.api_username = user['username']
        with pytest.raises(discourse.DiscourseClientForbidden):
            i.discourse.d._get(f'/c/{i.name_get()}/show.json')
    finally:
        i.discourse.d.api_username = api_username
    return True


def verify_moderation_is_no_archive_is_public(i, nonmember, member, moderator):
    subject, topic_id = email_stranger_ok(i, i.email_get())
    assert topic_id == user_can_see_topic(i, nonmember, subject)


def verify_moderation_is_no_archive_is_private(i, nonmember, member, moderator):
    subject, topic_id = email_stranger_ok(i, i.email_get())
    assert user_cannot_see_category(i, nonmember)
    assert topic_id == user_can_see_topic(i, member, subject)


def verify_moderation_is_ignore_archive_is_public(i, nonmember, member, moderator):
    topic_post_forbidden(i, nonmember)
    title, topic_id = topic_post_ok(i, member)
    email_stranger_ignored(i, i.email_get())
    assert topic_id == user_can_see_topic(i, nonmember, title)


def verify_moderation_is_ignore_archive_is_private(i, nonmember, member, moderator):
    assert user_cannot_see_category(i, nonmember)
    title, topic_id = topic_post_ok(i, member)
    assert topic_id == user_can_see_topic(i, member, title)
    email_stranger_ignored(i, i.email_get())


def verify_moderation_is_approval_archive_is_public(i, nonmember, member, moderator):
    title, topic_id = email_stranger_and_approve(i, i.email_get())
    assert topic_id == user_can_see_topic(i, nonmember, title)
    title, topic_id = topic_post_and_approve(i, member)
    assert topic_id == user_can_see_topic(i, nonmember, title)


def verify_moderation_is_approval_archive_is_private(i, nonmember, member, moderator):
    assert user_cannot_see_category(i, nonmember)
    title, topic_id = email_stranger_and_approve(i, i.email_get())
    assert topic_id == user_can_see_topic(i, member, title)
    title, topic_id = topic_post_and_approve(i, member)
    assert topic_id == user_can_see_topic(i, member, title)


@pytest.mark.parametrize("info,verify", [
    ({'moderation': 'no', 'archive': 'public'},
     verify_moderation_is_no_archive_is_public),
    ({'moderation': 'no', 'archive': 'private'},
     verify_moderation_is_no_archive_is_private),
    ({'moderation': 'ignore', 'archive': 'public'},
     verify_moderation_is_ignore_archive_is_public),
    ({'moderation': 'ignore', 'archive': 'private'},
     verify_moderation_is_ignore_archive_is_private),
    ({'moderation': 'approval', 'archive': 'public'},
     verify_moderation_is_approval_archive_is_public),
    ({'moderation': 'approval', 'archive': 'private'},
     verify_moderation_is_approval_archive_is_private),
])
def test_importer_category_settings(test_options, info, verify):
    i = importer.Importer(test_options)
    owner = 'owner@example.com'
    moderator = 'moderator@example.com'
    member = 'member@example.com'
    nonmember = 'nonmember@example.com'
    try:
        c = i.mailman.load_pickle()
        mailman_add(c, owner, owner=True)
        mailman_add(c, moderator, moderator=True)
        mailman_add(c, member)
        mailman_write_config(test_options.mailman_config, c)
        i.mailman.load()
        i.mailman.info.update(info)

        password = 'dashkeedfakojfiesMob'
        username = 'NONMEMBERNAME'
        _, _ = i.discourse.user_create(nonmember, **{
            'active': 'true', 'username': username, 'password': password})

        _, _ = i.importer()

        verify(i, nonmember, member, moderator)

    finally:
        i.discourse.category_topics_delete(i.name_get())
        i.discourse.category_delete(i.name_get())
        i.discourse.group_delete(i.group_moderators_name())
        i.discourse.group_delete(i.group_members_name())
        i.discourse.user_delete(moderator)
        i.discourse.user_delete(owner)
        i.discourse.user_delete(member)
        i.discourse.user_delete(nonmember)


def test_importer_notifications_level(test_options):
    i = importer.Importer(test_options)
    member = 'member@example.com'
    member_no_mail = 'membernomail@example.com'
    try:
        c = i.mailman.load_pickle()
        mailman_add(c, member, delivery_status=0)
        mailman_add(c, member_no_mail, delivery_status=1)
        mailman_write_config(test_options.mailman_config, c)
        i.mailman.load()

        before0, after0 = i.importer()
        assert before0['category'] is None

        assert (after0['notifications'][member] ==
                Discourse.CATEGORY_NOTIFICATION_LEVEL_WATCHING)
        assert (after0['notifications'][member_no_mail] ==
                Discourse.CATEGORY_NOTIFICATION_LEVEL_MUTED)

        before1, after1 = i.importer()
        assert before1 == after0
        assert before1 == after1
    finally:
        i.discourse.group_delete(i.group_moderators_name())
        i.discourse.category_delete(i.name_get())
        i.discourse.user_delete(member)
        i.discourse.user_delete(member_no_mail)


@pytest.mark.parametrize("private_roster", [
    Mailman.ROSTER_PUBLIC,
    Mailman.ROSTER_MEMBERS,
])
def test_importer_user_preferences(test_options, private_roster):
    i = importer.Importer(test_options)
    fr_language = 'fr@example.com'
    no_language = 'none@example.com'
    try:
        c = i.mailman.load_pickle()
        c['private_roster'] = private_roster
        mailman_add(c, fr_language, language='fr')
        mailman_add(c, no_language)
        del c['language'][no_language]
        mailman_write_config(test_options.mailman_config, c)
        i.mailman.load()

        before0, after0 = i.importer()

        assert after0['preferences'][fr_language]['locale'] == 'fr'
        if private_roster == Mailman.ROSTER_MEMBERS:
            assert after0['preferences'][no_language]['locale'] == ''
        else:
            assert no_language not in after0['preferences']

        before1, after1 = i.importer()
        assert before1 == after0
        assert before1 == after1
    finally:
        i.discourse.group_delete(i.group_moderators_name())
        i.discourse.category_delete(i.name_get())
        i.discourse.user_delete(fr_language)
        i.discourse.user_delete(no_language)


def test_importer_category_info(test_options):
    i = importer.Importer(test_options)
    try:
        long_description = 'This is a description'
        c = i.mailman.load_pickle()
        c['description'] = long_description
        mailman_write_config(test_options.mailman_config, c)
        i.mailman.load()

        before0, after0 = i.importer()
        assert after0['category-info'] == long_description
    finally:
        i.discourse.category_delete(i.name_get())
